// --------------------------------------------------------------------------
// parallel.c
// A program that reads command lines and executes them by a given number
// of processes in parallel.
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2013, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// -------------------------------------------------------------------------

#define _POSIX_C_SOURCE 199309L

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <pthread.h>

// --------------------------------------------------------------------------
// the data
// --------------------------------------------------------------------------

// a task description to be processed by a thread
typedef struct
{
  pthread_t       thread;  // the thread executing the task
  pthread_mutex_t mutex;   // the mutex associated to the thread
  pthread_cond_t  cond;    // the condition associated to the thread
  int             counter; // a unique identifier for the task
  char*           command; // the command to be executed
  uint64_t        time;    // the time for executing the command in ns
  bool            ready;   // indication that thread is ready to accept task
} task_t;

// the result delivered by a task
static task_t *taskResult;
static pthread_mutex_t resultMutex;
static pthread_cond_t fullCond;
static pthread_cond_t emptyCond;

// a command to be executed
typedef struct command_t 
{
  char*  line;
  struct command_t* next;
} command_t;

// the queue of commands to be executed
static command_t* commandsHead;
static command_t* commandsTail;
static int commandsNumber;
static pthread_cond_t commandsCond;
static pthread_mutex_t commandsMutex;

// a buffer for reading lines of text
static char* buffer;
static int bufferLength;
static int bufferPos;

// true if eof encountered
static bool eof;

// initialize the data
static void initialize(void)
{
  pthread_mutex_init(&resultMutex, NULL);
  pthread_cond_init(&fullCond, NULL);
  pthread_cond_init(&emptyCond, NULL);
  commandsHead = NULL;
  commandsTail = NULL;
  commandsNumber = 0;
  pthread_cond_init(&commandsCond, NULL);
  pthread_mutex_init(&commandsMutex, NULL);
  bufferLength = 512;
  bufferPos = 0;
  buffer = (char*)malloc(bufferLength*sizeof(char));
  eof = false;
}

// finalize the data
static void finalize()
{
  free(buffer);
}

// --------------------------------------------------------------------------
// the main program
// --------------------------------------------------------------------------

static void* read(void *arg);
static void* execute(void *command);
static char* getLine(void);
static uint64_t timespecDiff(struct timespec *timeA_p, 
  struct timespec *timeB_p);

static void usage(void)
{
  printf(
    "Usage: parallel <PROC>\n"
    "  PROC ... the number of parallel processes (>=1)\n\n"
    "  This command reads from its standard input (also from a pipe)\n" 
    "  arbitrarily (also infinitely) many command lines and executes\n"
    "  them by the denoted number of processes in parallel.\n\n"
    "  Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>\n"
    "  Copyright (C) 2013, Research Institute for Symbolic Computation (RISC)\n"
    "  Johannes Kepler University, Linz, Austria, http://www.risc.jku.at\n\n"
    "  This program is free software under the terms of the\n"
    "  GNU General Public License (GPL) version 3 or later.\n\n"
    "  The source code of this program can be downloaded from\n"
    "  http://www.risc.jku.at/people/schreine/parallel/parallel.tgz\n");
  exit(-1);
}

// the main function which manages the tasks
int main(int argc, char** argv)
{
  // process arguments
  if (argc != 2) usage();
  int taskLength = atoi(argv[1]);
  if (taskLength < 1) usage();

  // initialize data
  initialize();

  // create thread for handling input (terminates if there is no more input)
  pthread_t rthread;
  int result = pthread_create(&rthread, NULL, read, NULL);
  if (result != 0)
  {
    printf("could not create read thread (%d)\n", result);
    exit(-1);
  }

  // create worker threads
  task_t* tasks = (task_t*)malloc(taskLength*sizeof(task_t));
  int taskNumber = 0;
  int taskCounter = 0;
  int openNumber = 0;
  bool done = false;
  int i;
  for (i=0; i<taskLength; i++)
  {
    pthread_mutex_init(&tasks[i].mutex, NULL);
    pthread_cond_init(&tasks[i].cond, NULL);
    char* command = getLine();
    done = command == NULL;
    if (done) break;
    tasks[i].counter = taskCounter;
    tasks[i].command = command;
    tasks[i].ready = false;
    taskCounter++; taskNumber++; openNumber++;
    result = pthread_create(&tasks[i].thread, NULL, execute, tasks+i);
    if (result != 0)
    {
      printf("could not create worker thread (%d)\n", result);
      exit(-1);
    }
  }

  // process tasks
  while (openNumber > 0)
  {
    // wait for any result
    pthread_mutex_lock(&resultMutex);
    while (taskResult == NULL) 
      pthread_cond_wait(&fullCond, &resultMutex);
    task_t* tresult = taskResult;
    taskResult = NULL;
    pthread_cond_signal(&emptyCond);
    pthread_mutex_unlock(&resultMutex);
    openNumber--;

    // print result information
    printf("\n");
    printf("task:      %d\n", tresult->counter);
    printf("command:   %s\n", tresult->command);
    printf("time (ms): %lu\n", tresult->time/1000000);
    printf("executing: %d\n", openNumber);
    printf("queued:    %d\n", commandsNumber);
    fflush(stdout);
    free(tresult->command);
    tresult->command = NULL;

    // get new task
    if (done) continue;
    char* command = getLine();
    done = command == NULL;
    if (done) continue;
    
    // let thread execute the new task
    pthread_mutex_lock(&tresult->mutex);
    tresult->counter = taskCounter;
    tresult->command = command;
    tresult->ready = false;
    pthread_cond_signal(&tresult->cond);
    pthread_mutex_unlock(&tresult->mutex);
    taskCounter++; openNumber++; 
  }

  // terminate threads
  for (i=0; i<taskNumber; i++)
  {
    result = pthread_cancel(tasks[i].thread);
    if (result != 0)
    {
      printf("could not cancel thread (%d)\n", result);
      exit(-1);
    }
    result = pthread_join(tasks[i].thread, NULL);
    if (result != 0)
    {
      printf("could not join canceled thread (%d)\n", result);
      exit(-1);
    }
  }
  result = pthread_join(rthread, NULL);
  if (result != 0)
  {
    printf("could not join reader thread (%d)\n", result);
    exit(-1);
  }

  // finalize data
  finalize();
  return 0;
}

// forever execute assigned commands
static void* execute(void *arg) 
{
  task_t* task = (task_t*)arg;
  while(true)
  {
    // execute the assigned command
    struct timespec start, end;
    clock_gettime(CLOCK_MONOTONIC, &start);
    int result = system(task->command);
    clock_gettime(CLOCK_MONOTONIC, &end);
    task->time = timespecDiff(&end, &start);
    if (result == -1)
    {
      printf("could not execute: %s\n", task->command);
      exit(-1);
    }

    // lock access to task (*before* delivering result)
    pthread_mutex_lock(&task->mutex);
        
    // deliver result
    pthread_mutex_lock(&resultMutex);
    while (taskResult != NULL)
      pthread_cond_wait(&emptyCond, &resultMutex);
    taskResult = task;
    pthread_cond_signal(&fullCond);
    pthread_mutex_unlock(&resultMutex);
    
    // get next assignment
    task->ready = true;
    do
    {
      pthread_cond_wait(&task->cond, &task->mutex);
    }
    while (task->ready);
    
    // unlock access to task
    pthread_mutex_unlock(&task->mutex);
  }
}

// --------------------------------------------------------------------------
// time handling
// --------------------------------------------------------------------------

// compute the difference of times in nanoseconds
static uint64_t timespecDiff(struct timespec *timeA_p, struct timespec *timeB_p)
{
  return (timeA_p->tv_sec * 1000000000 + timeA_p->tv_nsec) -
         (timeB_p->tv_sec * 1000000000 + timeB_p->tv_nsec);
}

// --------------------------------------------------------------------------
// input handling
// --------------------------------------------------------------------------

static char* readLine(void);
static void addLine(char* line);
static void noMoreLine(void);

// read input into command buffer
static void* read(void *arg __attribute__((unused)))
{
  while (true)
  {
    char* line = readLine();
    if (line == NULL) break;
    addLine(line);
  }
  noMoreLine();
  return NULL;
}

// add line to commands queue
static void addLine(char* line)
{
  command_t* command = (command_t*)malloc(sizeof(command_t));
  command->line = line;
  command->next = NULL;
  pthread_mutex_lock(&commandsMutex);
  if (commandsTail == NULL)
    commandsHead = command;
  else
    commandsTail->next = command;
  commandsTail = command;
  commandsNumber++;
  pthread_cond_signal(&commandsCond);
  pthread_mutex_unlock(&commandsMutex);
}

// indicate that there is no more line
static void noMoreLine(void)
{
  pthread_mutex_lock(&commandsMutex);
  pthread_cond_signal(&commandsCond);
  pthread_mutex_unlock(&commandsMutex);
}

// get line from commands array
static char* getLine(void)
{
  pthread_mutex_lock(&commandsMutex);
  while (commandsHead == NULL)
  {
    if (eof)
    {
      pthread_mutex_unlock(&commandsMutex);
      return NULL;
    }
    pthread_cond_wait(&commandsCond, &commandsMutex);
  }
  char* result = commandsHead->line;
  command_t* head = commandsHead;
  commandsHead = commandsHead->next;
  free(head);
  if (commandsHead == NULL) commandsTail = NULL;
  commandsNumber--;
  pthread_mutex_unlock(&commandsMutex);
  return result;
}

static void addChar(char ch);

// read line from standard input and return it
static char* readLine(void)
{
  if (eof) return NULL;
  while (true)
  {
    int ch = getchar();
    if (ch == EOF)  
    { 
      eof = true;
      if (bufferPos == 0) return NULL; 
      break; 
    }
    if (ch == '\n') break;
    addChar((char)ch);
  }
  char *line = (char*)malloc((bufferPos+1)*sizeof(char));
  memcpy(line, buffer, bufferPos*sizeof(char));
  line[bufferPos] = 0;
  bufferPos = 0;
  return line;
}

// add character to buffer
static void addChar(char ch)
{
  if (bufferPos == bufferLength)
  {
    int bufferLength0 = 2*bufferLength;
    char *buffer0 = (char*)malloc(bufferLength0*sizeof(char));
    memcpy(buffer0, buffer, bufferLength*sizeof(char));
    free(buffer);
    buffer = buffer0;
    bufferLength = bufferLength0;
  }
  buffer[bufferPos] = ch;
  bufferPos++;
}

// --------------------------------------------------------------------------
// end of file
// --------------------------------------------------------------------------

